#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import glob
import traceback
import lxml.etree
import argparse
from foliatools import VERSION as TOOLVERSION
import folia.main as folia


def validate(filename, schema = None,**kwargs):
    try:
        folia.validate(filename, schema)
    except Exception as e:
        print("VALIDATION ERROR against RelaxNG schema (stage 1/2), in " + filename,file=sys.stderr)
        print(str(e), file=sys.stderr)
        return False
    try:
        document = folia.Document(file=filename, deepvalidation=kwargs.get('deep',False),textvalidation=kwargs.get('stricttextvalidation',False),verbose=True, autodeclare=kwargs.get('autodeclare',False), debug=kwargs.get('debug'))
    except folia.DeepValidationError as e:
        print("DEEP VALIDATION ERROR on full parse by library (stage 2/2), in " + filename,file=sys.stderr)
        print(e.__class__.__name__ + ": " + str(e),file=sys.stderr)
        return False
    except Exception as e:
        print("VALIDATION ERROR on full parse by library (stage 2/2), in " + filename,file=sys.stderr)
        print(e.__class__.__name__ + ": " + str(e),file=sys.stderr)
        if kwargs.get('traceback') or kwargs.get('debug'):
            print("-- Full traceback follows -->",file=sys.stderr)
            ex_type, ex, tb = sys.exc_info()
            traceback.print_exception(ex_type, ex, tb)
        return False
    if not document.version:
        print("VALIDATION ERROR: Document does not advertise FoLiA version (" + filename + ")",file=sys.stderr)
        return False
    elif folia.checkversion(document.version) == -1 and not kwargs.get('nowarn'):
        if document.version.split('.')[0] in ('0','1'):
            print("WARNING: Document (" + filename + ") uses an older FoLiA version ("+document.version+") but is validated with a newer library (" + folia.FOLIAVERSION+"). If this is a document you created and intend to publish, you may want to upgrade this FoLiA v1 document to FoLiA v2 using the 'foliaupgrade' tool.",file=sys.stderr)
        else:
            print("WARNING: Document (" + filename + ") uses an older FoLiA version ("+document.version+") but is validated according to the newer specification (" + folia.FOLIAVERSION+"). You might want to increase the version attribute if this is a document you created and intend to publish.",file=sys.stderr)
    if document.textvalidationerrors:
        if kwargs.get('stricttextvalidation'):
            print("VALIDATION ERROR because of text validation errors, in " + filename,file=sys.stderr)
            return False
        elif not kwargs.get('nowarn'):
            print("WARNING: there were " + str(document.textvalidationerrors) + " text validation errors but these are currently not counted toward the full validation result (use -t for strict text validation)", file=sys.stderr)

    if kwargs.get('output'):
        print(document.xmlstring())
    if kwargs.get('autodeclare'):
        print("Validated successfully **after** applying auto-declarations: " +  filename,file=sys.stderr)
        return document
    else:
        print("Validated successfully: " +  filename,file=sys.stderr)
        return True




def processdir(d, schema = None, **kwargs):
    success = False
    print("Searching in  " + d,file=sys.stderr)
    extension = kwargs.get('extension','xml').strip('.')
    for f in glob.glob(os.path.join(d ,'*')):
        r = True
        if f[-len(extension) - 1:] == '.' + extension:
            r = validate(f, schema,**kwargs)
        elif kwargs.get('recurse') and os.path.isdir(f):
            r = processdir(f,schema,**kwargs)
        if not r: success = False
    return success

def commandparser(parser):
    parser.add_argument('-d','--deep',help="Enable deep validation; validated uses classes against provided set definitions", action='store_true', default=False)
    parser.add_argument('-r','--recurse',help="Process recursively", action='store_true', default=False)
    parser.add_argument('-a','--autodeclare',help="Attempt to automatically declare missing annotations", action='store_true', default=False)
    parser.add_argument('-q','--quick',help="Quick and more shallow validation, only validates against RelaxNG schema. This does not constitute a complete enough validation!", action='store_true', default=False)
    parser.add_argument('-E','--extension', type=str,help="Extension", action='store',default="xml")
    parser.add_argument('-W','--nowarn',help="Suppress warnings", action='store_true', default=False)
    parser.add_argument('-i','--ignore',help="Always report a successful exit code, even in case of validation errors", action='store_true', default=False)
    parser.add_argument('-t','--stricttextvalidation',help="Treat text validation errors strictly (recommended and default for FoLiA v1.5+)", action='store_true', default=False)
    parser.add_argument('-o','--output',help="Output document to stdout", action='store_true', default=False)
    parser.add_argument('-D','--debug',type=int,help="Debug level", action='store',default=0)
    parser.add_argument('-b','--traceback',help="Provide a full traceback on validation errors", action='store_true', default=False)
    return parser

def main():
    parser = argparse.ArgumentParser(description="Checks whether a FoLiA document is a valid FoLiA document, i.e. whether it properly adheres to the FoLiA specification. Invalid documents should never be used or published.", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-v','-V','--version',help="Show version information", action='version', version="FoLiA-tools v" + TOOLVERSION + ", using FoLiA v" + folia.FOLIAVERSION + " with library FoLiApy v" + folia.LIBVERSION, default=False)
    commandparser(parser)
    parser.add_argument('files', nargs='*', help='Files (and/or directories) to validate')
    args = parser.parse_args()

    schema  = lxml.etree.RelaxNG(folia.relaxng())

    if args.files:
        success = True
        skipnext = False
        for file in args.files:
            r = False
            if os.path.isdir(file):
                r = processdir(file,schema, **args.__dict__)
            elif os.path.isfile(file):
                r = validate(file, schema, **args.__dict__)
            else:
                print("ERROR: File or directory not found: " + file,file=sys.stderr)
                sys.exit(3)
            if not r: success= False
        if not success and not args.ignore:
            sys.exit(1)
    else:
        print("ERROR: No files specified. Add --help for usage details.",file=sys.stderr)
        sys.exit(2)

if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf8 -*-

import getopt
import sys
from foliatools import VERSION
import folia.main as folia

def usage():
    print("foliamerge",file=sys.stderr)
    print("  by Maarten van Gompel (proycon)",file=sys.stderr)
    print("  Radboud University Nijmegen",file=sys.stderr)
    print("  2017-2019 - Licensed under GPLv3",file=sys.stderr)
    print("",file=sys.stderr)
    print("Merges annotations from two or more FoLiA documents. Structural elements are never added. Annotations can only be merged if their parent elements have IDs.",file=sys.stderr)
    print("",file=sys.stderr)
    print("Usage: foliamerge [options] file1 file2 file3 ... ",file=sys.stderr)
    print("",file=sys.stderr)
    print("Options:",file=sys.stderr)
    print("  -o [file]                    Output file",file=sys.stderr)
    print("  -s                           Substitute: use first input file as output as well",file=sys.stderr)
    print("  -a                           Annotations from file2 onwards are included as alternatives",file=sys.stderr)

def attach(parent, child):
    """Rather than copy children, we just abduct them from the other-doc parent"""
    child.doc = parent.doc
    child.parent = parent
    child.setparents()

def reID(doc, element, oldprefix, newprefix):
    if element.id:
        element.id = element.id.replace(oldprefix + '.', newprefix + '.')
    for child in element.data:
        if isinstance(child, folia.AbstractElement) and not isinstance(child, folia.AbstractStructureElement):
            reID(doc, child, oldprefix, newprefix)

def mergechildren(parent, outputdoc, asalternative):
    if hasattr(parent,'merged'): return 0
    merges = 0
    for e in parent:
        if isinstance(e, (folia.AbstractTokenAnnotation, folia.AbstractAnnotationLayer)) and parent.id and not hasattr(e, 'merged'):
            try:
                e.ANNOTATIONTYPE
            except:
                continue

            if (e.ANNOTATIONTYPE, e.set) in outputdoc.annotations:
                assert e.parent == parent
                try:
                    newparent = outputdoc[parent.id]
                except:
                    continue #parent does not exist, nothing to merge here
                if isinstance(newparent, (folia.Alternative, folia.AlternativeLayers)):
                    #we do not merge alternatives, each alternative has its own scope
                    continue
                #check if the annotation already exists
                #print("DEBUG: Adding annotation type " + e.__class__.__name__ + ", set " + e.set + ", under " + newparent.id, file=sys.stderr)
                if asalternative:
                    if isinstance(e, folia.AbstractTokenAnnotation):
                        print("Adding Annotation type " + e.__class__.__name__ + ", set " + str(e.set) + " to " + newparent.id + " as alternative", file=sys.stderr)
                        alt = newparent.append(folia.Alternative, generate_id_in=newparent)
                        reID(newparent.doc, e, newparent.id, alt.id)
                        alt.append(e)
                        e.setdoc(outputdoc)
                        assert e.parent is alt
                        alt.merged = True
                        e.merged = True
                        merges += 1
                    elif isinstance(e, folia.AbstractAnnotationLayer):
                        print("Adding Annotation type " + e.__class__.__name__ + ", set " + str(e.set) + " to " + newparent.id + " as alternative", file=sys.stderr)
                        alt = newparent.append(folia.AlternativeLayers, generate_id_in=newparent)
                        reID(newparent.doc, e, newparent.id, alt.id)
                        alt.append(e)
                        assert e.parent is alt
                        e.setdoc(outputdoc)
                        alt.merged = True
                        e.merged = True
                        merges += 1
                elif isinstance(e, folia.AbstractTokenAnnotation) and newparent.hasannotation(e.__class__, e.set):
                    print("Annotation type " + e.__class__.__name__ + ", set " + e.set + ", under " + newparent.id + " , already exists... skipping", file=sys.stderr)
                    pass
                elif isinstance(e, folia.AbstractAnnotationLayer) and newparent.hasannotationlayer(e.__class__, e.set):
                    print("Annotation type " + e.__class__.__name__ + ", set " + e.set + ", under " + newparent.id + " , already exists... skipping", file=sys.stderr)
                    pass
                else:
                    print("Adding Annotation type " + e.__class__.__name__ + ", set " + str(e.set) + " to " + newparent.id, file=sys.stderr)
                    newparent.append(e)
                    assert e.parent is newparent
                    e.setdoc(outputdoc)
                    e.merged = True
                    merges += 1
        elif isinstance(e, folia.AbstractElement):
            merges += mergechildren(e, outputdoc, asalternative)
    return merges



def foliamerge(outputfile, *files, **kwargs):
    asalternative = 'asalternative' in kwargs and kwargs['asalternative']
    outputdoc = None
    merges = 0

    for i, filename in enumerate(files):
        print("Processing " + filename, file=sys.stderr)
        inputdoc = folia.Document(file=filename)
        if i == 0:
             print("(pivot document)",file=sys.stderr)
             outputdoc = inputdoc
        else:
            print("(merging document)",file=sys.stderr)

            for annotationtype,set in inputdoc.annotations:
                if not outputdoc.declared(annotationtype,set):
                    outputdoc.declare( annotationtype, set)

            for e in inputdoc:
                merges += mergechildren(e, outputdoc, asalternative)

    if outputfile and merges > 0:
        outputdoc.provenance.append(folia.Processor.create(name="foliamerge", version=VERSION))
        outputdoc.save(outputfile)

    return outputdoc

def main():
    try:
        opts, args = getopt.getopt(sys.argv[1:], "o:sha", ["help"])
    except getopt.GetoptError as err:
        print(str(err),file=sys.stderr)
        usage()
        sys.exit(2)

    outputfile = None
    substitute = False
    asalternative = False

    for o, a in opts:
        if o == '-h' or o == '--help':
            usage()
            sys.exit(0)
        elif o == '-o':
            outputfile = a
        elif o == '-s':
            substitute = True
        elif o == '-a':
            asalternative = True
        else:
            raise Exception("No such option: " + o)

    if len(args) < 2:
        print("ERROR: At least two files need to be specified",file=sys.stderr)
        sys.exit(2)

    if substitute:
        outputfile = args[0]

    outputdoc = foliamerge(outputfile, *args, asalternative=asalternative)
    if not outputfile:
        xml = outputdoc.xmlstring()
        print(xml)


if __name__ == "__main__":
    main()






#!/usr/bin/env python
#-*- coding:utf-8 -*-

#---------------------------------------------------------------
# CONNL-U to FoLiA converter
#   by Maarten van Gompel
#   Centre for Language and Speech technology
#   Radboud University Nijmegen
#   proycon AT anaproy DOT nl
#
#   Licensed under GPLv3
#
# This script converts CONNL-U to FoLiA format.
#
#----------------------------------------------------------------

from __future__ import print_function, unicode_literals, division, absolute_import

import sys
import os
import argparse
import conllu
from pynlpl.formats import folia

UPOS_SET = "https://raw.githubusercontent.com/proycon/folia/master/setdefinitions/universal-pos.foliaset.ttl"
UDEP_SET = "https://raw.githubusercontent.com/proycon/folia/master/setdefinitions/universal-dependencies.foliaset.ttl"


def main():
    parser = argparse.ArgumentParser(description="CONLL-U to FoLiA converter", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--id',type=str,help="Document ID for the FoLiA document", action='store',default="",required=False)
    parser.add_argument('--lemma-set',dest="lemmaset", type=str,help="URL of the set definition for lemmas", action='store',default="undefined",required=False)
    parser.add_argument('--pos-set',dest="posset",type=str,help="URL of the set definition for *language-specific* part-of-speech and features (xpos and not the universal pos!)", action='store',default="undefined",required=False)
    parser.add_argument('--dependency-set',dest="depset", type=str,help="Dependency set", action='store',default=UDEP_SET, required=False)
    parser.add_argument('-o', '--outputdir',type=str,help="Output directory", action='store',default=".", required=False)
    parser.add_argument('files', nargs='+', help='CONLL-U input files')
    args = parser.parse_args()


    for file in args.files:
        if args.id:
            doc_id = args.id
        else:
            doc_id = os.path.basename(file)
        doc = None
        hascontent = False
        with open(file,'r',encoding='utf-8') as f:
            sentences = conllu.parse(f.read())
            for i, tokenlist in enumerate(sentences):
                if 'newdoc id' in tokenlist.metadata or i == 0:
                    if doc is not None and hascontent:
                        doc.save(os.path.join(args.outputdir, doc_id + ".folia.xml"))
                        print("Wrote " + doc_id + ".folia.xml",file=sys.stderr)
                    if 'newdoc id' in tokenlist.metadata:
                        doc_id = tokenlist.metadata['newdoc id']
                    hascontent = False
                    doc = folia.Document(id=doc_id)
                    doc.declare(folia.PosAnnotation, set=UPOS_SET, annotator="conll2folia")
                    doc.declare(folia.PosAnnotation, set=args.posset, annotator="conll2folia")
                    doc.declare(folia.Dependency, set=args.depset, annotator="conll2folia")
                    doc.declare(folia.LemmaAnnotation, set=args.lemmaset, annotator="conll2folia")
                    textbody = folia.Text(doc, id=doc_id+'.text')
                    doc.append(textbody)
                    anchor = textbody
                if 'newpar id' in tokenlist.metadata:
                    anchor = textbody.append(folia.Paragraph, id=tokenlist.metadata['newpar id'])
                if 'sent_id' in tokenlist.metadata:
                    sent_id = tokenlist.metadata['sent_id']
                else:
                    sent_id = doc_id + '.s.' + str(i+1)
                sentence = folia.Sentence(doc, id=sent_id)
                if 'text' in tokenlist.metadata:
                    sentence.add(folia.TextContent, tokenlist.metadata['text'], cls="original")
                elif 'text_en' in tokenlist.metadata:
                    sentence.add(folia.TextContent, tokenlist.metadata['text_en'], cls="text_en")
                wordindex = {} #quick lookup index for this sentence
                for token in tokenlist:
                    if token['misc'] and 'SpaceAfter' in token['misc'] and token['misc']['SpaceAfter'].lower() == 'no':
                        space = False
                    else:
                        space = True
                    word = sentence.add(folia.Word, token['form'], id=sent_id + ".w." + str(token['id']), space=space)
                    wordindex[token['id']] = word
                    if token['upostag']:
                        pos = word.add(folia.PosAnnotation, cls=token['upostag'], set=UPOS_SET)
                        if isinstance(token['feats'], dict):
                            for subset, cls in token['feats'].items():
                                pos.add(folia.Feature,subset=subset,cls=cls)
                    if token['xpostag']:
                        word.append(folia.PosAnnotation, cls=token['xpostag'], set=args.posset)
                        if isinstance(token['feats'], dict) and not token['upostag']:
                            for subset, cls in token['feats'].items():
                                pos.add(folia.Feature,subset=subset,cls=cls)
                    if token['lemma']:
                        word.add(folia.LemmaAnnotation, cls=token['lemma'], set=args.lemmaset)
                    hascontent = True
                for token in tokenlist:
                    if token['head'] and token['deprel']:
                        sentence.add(folia.Dependency(doc, set=args.depset, cls=token['deprel'], contents=[
                            folia.DependencyHead(doc, wordindex[token['head']]),
                            folia.DependencyDependent(doc, wordindex[token['id']])
                        ]))
                anchor.add(sentence)

        if doc is not None and hascontent:
            doc.save(os.path.join(args.outputdir, doc_id + ".folia.xml"))
            print("Wrote " + doc_id + ".folia.xml",file=sys.stderr)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
#Generate library specification code (for either Python or C++) on the the basis of folia.yml
#Used by respectively pynlpl and libfolia

from __future__ import print_function, unicode_literals, division, absolute_import

import sys
import os
import json
import yaml

#Load specification
specfiles= [  os.path.join(os.path.dirname(__file__) ,'../schemas/folia.yml'), 'folia.yml' ]
spec = None
for specfile in specfiles:
    spec = yaml.load(open(specfile,'r'))
    break

if spec is None:
    print("FoLiA Specification file folia.yml could not be found in " + ", ".join(specfiles) ,file=sys.stderr)

def main(var=None):
    try:
        var = sys.argv[1]
        if var[0] == '-': var = None
    except:
        var = None
    if var:
        print(var + ' = ' + json.dumps(spec, sort_keys=True, indent=4) + ';')
    else:
        print(json.dumps(spec, sort_keys=True, indent=4))

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import sys
import os.path
import argparse
import glob
import shutil
import folia.main as folia
from foliatools import VERSION
from foliatools.foliavalidator import validate

def main():
    parser = argparse.ArgumentParser(description="", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-V', '--version',help="Output version information", action='store_true')
    parser.add_argument('-n', '--dryrun',help="Dry run, do not write files", action='store_true', default=False)
    parser.add_argument('-E','--extension', type=str,help="Extension", action='store',default="xml",required=False)
    parser.add_argument('files', nargs='+', help='Input files')
    args = parser.parse_args()

    if args.version:
        print("FoLiA " + folia.FOLIAVERSION + ", library version " + folia.LIBVERSION,file=sys.stderr)
        sys.exit(0)

    success = process(*args.files, **args.__dict__)
    sys.exit(0 if success else 1)

def process(*files, **kwargs):
    success = False
    for file in files:
        if os.path.isdir(file):
            r = process(list(glob.glob(os.path.join(file, "*." + kwargs['extension'] )) ), **kwargs)
            if r != 0:
                success = False
        elif os.path.isfile(file):
            doc = validate(file,schema=None, stricttextvalidation=True,autodeclare=True,output=False, warn=False, **kwargs)
            doc.provenance.append(folia.Processor.create(name="foliaupgrade", version=VERSION))
            if doc is not False:
                print("Upgrading " + doc.filename,file=sys.stderr)
                doc.version = folia.FOLIAVERSION #upgrading involves more than just bumping the number, but that is handled implicitly already by the library when reading the document
                if not kwargs.get('dryrun'):
                    doc.save(doc.filename + ".upgraded")
                    if not validate(file + ".upgraded",schema=None,stricttextvalidation=True,autodeclare=True,**kwargs):
                        print("Upgrade failed",file=sys.stderr)
                        success = False
                    else:
                        shutil.move(doc.filename + ".upgraded", file)
            else:
                success  = False
    return success

if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf8 -*-


from __future__ import print_function, unicode_literals, division, absolute_import

import argparse
import sys
from foliatools import VERSION as TOOLVERSION
import folia.main as folia


def concat(target, source):
    merges = 0
    for e in source:
        c = e.copy(target.doc)
        target.append(c)
        merges += 1
    return merges



def foliacat(id, outputfile, *files, keepversion=True):
    totalmerges = 0
    outputdoc = folia.Document(id=id,keepversion=keepversion)
    if keepversion:
        outputdoc.version = "0.0.0"
    text = outputdoc.append(folia.Text(outputdoc,id=id + ".text"))
    for i, filename in enumerate(files):
        merges = 0
        print("Processing " + filename, file=sys.stderr)
        inputdoc = folia.Document(file=filename,keepversion=True)
        #we take the version of the newest document
        if keepversion:
            if folia.checkversion(inputdoc.version, outputdoc.version) > 0:
                outputdoc.version = inputdoc.version
        print("(merging document)",file=sys.stderr)

        for annotationtype,set in inputdoc.annotations:
            if not outputdoc.declared(annotationtype,set):
                outputdoc.declare( annotationtype, set)

        for d in inputdoc.data:
            merges += concat(text, d)

        print("(merged " + str(merges) + " elements, with all elements contained therein)",file=sys.stderr)
        totalmerges += merges


    print("(TOTAL: merged " + str(totalmerges) + " elements, with all elements contained therein)",file=sys.stderr)
    if outputfile and merges > 0:
        outputdoc.save(outputfile)

    return outputdoc

def main():
    parser = argparse.ArgumentParser(description="Concatenates two or more FoLiA documents", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-v','--version',help="Show version information", action='version', version="FoLiA-tools v" + TOOLVERSION + ", using FoLiA v" + folia.FOLIAVERSION + " with library FoLiApy v" + folia.LIBVERSION, default=False)
    parser.add_argument('-i','--id',type=str, help="Set the ID for the output document", action='store', required=True)
    parser.add_argument('-o','--output',type=str, help="Output file (defaults to stdout if not set)", required=True)
    parser.add_argument('-u','--upgrade', help="Automatically upgrade FoLiA version to the latest version", action='store_true')
    parser.add_argument('files', nargs='*', help='Files to concatenate')
    args = parser.parse_args()

    if len(args.files) < 2:
        print("WARNING: only one file specified", file=sys.stderr)


    outputdoc = foliacat(args.id, args.output, *args.files, keepversion=not args.upgrade)
    if not args.output or args.output == '-':
        print(outputdoc.xmlstring())

if __name__ == "__main__":
    main()
